/*
 * Copyright (c) 2012 - 2013 Alexandre Tuleu
 *
 * This file is part of biorob-cpp.
 *
 *   biorob-cpp is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as
 *   published by the Free Software Foundation, either version 3 of the
 *   License, or (at your option) any later version.
 *
 *   biorob-cpp is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of both the GNU General Public
 *   License AND GNU Lesser General Public License along with
 *   biorob-cpp.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "Prompt.h"

#include <stdexcept>

#include <algorithm>

#include <sstream>
#include <iomanip>


namespace cli {
Prompt::Prompt(const Widget::Def & def)
	: Widget(def)
	, d_starty(0)
	, d_prompt()
	, d_history()
	, d_historyPtr(d_history.end())
	, d_result()
	, d_backupResult()
	, d_curp(d_result.begin())
	, d_callbacks()
	, d_externalCallbacks()
	, d_currentCharRepeated(false)
	, d_availableCompletions(){
	scrollok(Window(),true);

	setlocale(LC_ALL,"");



	d_callbacks[KEY_LEFT] =&Prompt::DecrementCursor;
	d_callbacks[KEY_RIGHT] = &Prompt::IncrementCursor;


	d_callbacks[KEY_UP] = &Prompt::HistoryUp;
	d_callbacks[KEY_DOWN] = &Prompt::HistoryDown;

	d_callbacks[KEY_BACKSPACE] = &Prompt::EraseAtCursor;



	d_callbacks[ERR] = &Prompt::Ignore;



	d_callbacks[1] = &Prompt::MoveCursorAtBegin;//C-a
	d_callbacks[5] = &Prompt::MoveCursorAtEnd;//C-e
	d_callbacks[11] = &Prompt::EraseToEOL;//C-K
	d_callbacks[539] = &Prompt::MoveCursorToPrevWord;//C-Left arrow
	d_callbacks[554] = &Prompt::MoveCursorToNextWord;//C-Right arrow
	d_callbacks['\t'] = &Prompt::Complete;

	keypad(Window(),true);

	wrefresh(Window());
}


Prompt::~Prompt(){}

void Prompt::Ignore(){
	beep();
}

const std::string & Prompt::GetLine(const std::string & p){
	d_currentCharRepeated = false;
	d_prompt = p;
	d_historyPtr = d_history.end();
	d_result.clear();
	d_result.reserve(500);
	int startx,starty;
	waddstr(Window(),d_prompt.c_str());
	wrefresh(Window());
	getyx(Window(),starty,startx);
	d_starty = (size_t)starty;
	d_curp = d_result.begin();

	int c,lastC(0);


	while(true){
		wrefresh(Window());
		c = wgetch(Window());
		if(c == lastC){
			d_currentCharRepeated = true;
		} else {
			d_currentCharRepeated = false;
		}
		lastC = c;
		CallbackByKey::const_iterator fi = d_callbacks.find(c);
		if(fi != d_callbacks.end()){
			(this->*(fi->second))();

			continue;
		}

		KeyCallbackByKey::const_iterator externalCallback = d_externalCallbacks.find(c);
		if(externalCallback != d_externalCallbacks.end()){
			externalCallback->second->Call(c);
			continue;
		}


		if(c == '\n' || c == '\r' || c == KEY_ENTER){
			MoveCursorAtEnd();
			waddch(Window(),'\n');
			break;
		}


		InsertCharAtCursor(c);

	}
	wrefresh(Window());

	if(!d_result.empty() && d_result.at(d_result.size() - 1) == ' '){
		d_result.erase(d_result.size() - 1);
	}

	if(d_history.empty() || d_history.back() != d_result){
		d_history.push_back(d_result);
	}

	return d_result;
}

void Prompt::MoveCursorAtEnd(){
	d_curp = d_result.end();
	MoveCursorAtInsertionPoint();
}

void Prompt::MoveCursorAtBegin(){
	d_curp = d_result.begin();
	MoveCursorAtInsertionPoint();
}

void Prompt::MoveCursorAtInsertionPoint(){
	int x,y;
	LineToWinPos((std::size_t)(d_curp - d_result.begin()) + d_prompt.size(),x,y);
	wmove(Window(), y, x);
}

void Prompt::MoveCursorToNextWord(){
	while(d_curp != d_result.end() && *d_curp != ' '){
		++d_curp;
	}
	if(d_curp != d_result.end()){
		++d_curp;
	}

	MoveCursorAtInsertionPoint();
}

void Prompt::MoveCursorToPrevWord(){
	while(d_curp != d_result.begin() && *d_curp != ' '){
		--d_curp;
	}
	if(d_curp != d_result.begin()){
		--d_curp;
	}
	MoveCursorAtInsertionPoint();
}


void Prompt::InsertCharAtCursor(int c, bool escape){
	if(!escape){
		d_curp = d_result.insert(d_curp,c);
		++d_curp;
	}

	if(d_curp == d_result.end()){
		waddch(Window(),(chtype)c);
		return;
	}

	RefreshLine();
}


void Prompt::IncrementCursor(){
	if(d_curp == d_result.end()){
		beep();
		return;
	}
	++d_curp;
	MoveCursorAtInsertionPoint();
}

void Prompt::DecrementCursor(){
	if(d_curp == d_result.begin()){
		beep();
		return;
	}
	--d_curp;
	MoveCursorAtInsertionPoint();
}

void Prompt::EraseAtCursor(){
	if(d_curp == d_result.begin()){
		beep();
		return;
	}



	d_curp = d_result.erase(d_curp-1,d_curp);
	int x,y;
	LineToWinPos(d_result.size() + d_prompt.size(),x,y);
	mvwdelch(Window(),y,x);

	RefreshLine();

}

void Prompt::EraseToEOL(){
	d_result.erase(d_curp,d_result.end());
	d_curp = d_result.end();
	wclrtobot(Window());
}


void Prompt::RefreshLine(){
	wmove(Window(),(int)d_starty,0);
	waddstr(Window(),d_prompt.c_str());
	waddstr(Window(),d_result.c_str());

	MoveCursorAtInsertionPoint();
}


void Prompt::LineToWinPos(size_t pos,int & x, int & y){
	x = (int) (pos % (std::size_t) Width());
	y = (int) (d_starty + pos / Width());
}


void Prompt::HistoryUp(){
	if(d_historyPtr == d_history.begin()){
		beep();
		return;
	}
	if(d_historyPtr == d_history.end()){
		d_backupResult = d_result;
	}

	--d_historyPtr;
	d_result = *d_historyPtr;

	d_curp = d_result.end();
	wmove(Window(),(int)d_starty,0);
	wclrtobot(Window());
	RefreshLine();
}

void Prompt::HistoryDown(){
	if(d_historyPtr == d_history.end()){
		beep();
		return;
	}

	++d_historyPtr;

	if(d_historyPtr == d_history.end()){
		d_result = d_backupResult;
	} else {
		d_result = *d_historyPtr;
	}

	d_curp = d_result.end();
	wmove(Window(),(int)d_starty,0);
	wclrtobot(Window());
	RefreshLine();

}


void Prompt::Complete(){
	std::string curWord;
	std::string::iterator startWord,endWord;
	GetCurrentWord(curWord,startWord,endWord);
	unsigned int resSize;
	CompletionResult res =GetCompletionResult(curWord,resSize);

	if(resSize == 0){
		beep();
		return;
	}

	if(d_currentCharRepeated){
		PrintCompletion(res);
		RefreshLine();
		return;
	}

	CompletionSet::const_iterator m = std::min_element(res.first,res.second);
	std::string toInsert = *m;

	if(resSize > 1){
		bool cont = true;
		for(std::size_t i = 1; cont && i <= toInsert.size(); ++i){
			for(CompletionSet::const_iterator c = res.first;
			    c != res.second;
			    ++c){
				if(c->substr(0,i) != toInsert.substr(0,i)){
					cont = false;
					break;
				}
			}
			if(cont == false){
				toInsert.erase(i - 1,std::string::npos);
			}
		}
	}


	if(toInsert != curWord && !curWord.empty()){
		d_curp = d_result.erase(startWord,endWord);
		std::size_t pos = (std::size_t)(d_curp - d_result.begin());
		if(resSize == 1){
			toInsert += " ";
		}
		d_result.insert(pos,toInsert);
		d_curp = d_result.begin() + (std::string::difference_type) (pos + toInsert.size());

	} else {
		beep();
	}

	RefreshLine();

}


void Prompt::GetCurrentWord(std::string & curWord,
                            std::string::iterator & startWord,
                            std::string::iterator & endWord){
	for(startWord = d_curp; startWord != d_result.begin(); --startWord){
		if(*(startWord - 1) == ' '){
			break;
		}
	}

	for(endWord = d_curp; endWord != d_result.end(); ++endWord){
		if(*endWord == ' '){
			break;
		}
	}

	curWord.clear();
	curWord.append(startWord,endWord);

}


Prompt::CompletionResult
Prompt::GetCompletionResult(const std::string & val,
                            unsigned int & size){
	CompletionResult res;
	size = 0;
	if(val.empty()){
		res.first = d_availableCompletions.begin();
		res.second = d_availableCompletions.end();
		size = d_availableCompletions.size();
		return res;
	}
	res.first = d_availableCompletions.lower_bound(val);
	for(res.second = res.first;
	    res.second != d_availableCompletions.end();
	    ++(res.second)){
		if(val != res.second->substr(0,val.size())){
			break;
		}
		++size;
	}
	return res;
}

void Prompt::PrintCompletion(CompletionResult & res){
	if(res.first == res.second){
		return;
	}
	size_t max_size(0);
	unsigned int nbResults(0);
	for(CompletionSet::const_iterator i = res.first;
	    i != res.second;
	    ++i){
		if(i->size() > max_size){
			max_size = i->size();
		}
		++nbResults;
	}

	unsigned int nbCols = Width() / (max_size + 1);
	unsigned int nbLines = nbResults / nbCols + 1;
	unsigned int nbSpaces = (Width() - nbCols * max_size ) / nbCols;
	unsigned int colSize = max_size + nbSpaces;
	bool insertEOL(nbCols * colSize < Width());

	std::ostringstream os;
	unsigned int i(0);
	for(CompletionSet::const_iterator c = res.first;
	    c != res.second;
	    ++c){
		os << std::left << std::setw((int)colSize) << *c;
		if( insertEOL && ((i % nbCols) == (nbCols - 1))){
			os << std::endl;
		}
		++i;
	}

	wmove(Window(),(int)d_starty,0);
	wprintw(Window(),"%s", os.str().c_str());
	if( (nbResults % nbCols ) != 0){
		waddch(Window(),'\n');
	} else {
		nbLines -= 1;
	}
	d_starty += nbLines;
	RefreshLine();
}


void Prompt::AddCompletion(const std::string & val){
	if(val.empty()){
		return;
	}

	d_availableCompletions.insert(val);
}


void Prompt::RemoveCompletion(const std::string & val){
	CompletionSet::iterator fi = d_availableCompletions.find(val);

	if(fi == d_availableCompletions.end()){
		return;
	}

	d_availableCompletions.erase(fi);
}


void Prompt::BindKey(int c, KeyCallback::Ptr callback){
	KeyCallbackByKey::const_iterator fi = d_externalCallbacks.find(c);
	if(fi != d_externalCallbacks.end()){
		return;
	}

	d_externalCallbacks[c] = callback;
}

void Prompt::UnBindKey(int c){
	KeyCallbackByKey::iterator fi = d_externalCallbacks.find(c);
	if(fi == d_externalCallbacks.end()){
		return;
	}
	d_externalCallbacks.erase(fi);
}

}// namespace cli
