/*
 * Copyright (c) 2012 - 2013 Alexandre Tuleu
 *
 * This file is part of biorob-cpp.
 *
 *   biorob-cpp is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as
 *   published by the Free Software Foundation, either version 3 of the
 *   License, or (at your option) any later version.
 *
 *   biorob-cpp is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of both the GNU General Public
 *   License AND GNU Lesser General Public License along with
 *   biorob-cpp.  If not, see <http://www.gnu.org/licenses/>.
 */


/**
 * \file Config.h
 *
 *  \date Dec 3, 2012
 *  \author Alexandre Tuleu
 */

#ifndef BIOROB_CPP_OPTIONS_CONFIG_H_
#define BIOROB_CPP_OPTIONS_CONFIG_H_

#include "Option.h"
#include <iostream>
#include <stdexcept>

#define BCPP_ADD_OPTION(Name,NameInFile,Type,DefaultValue,Comment) \
	private :\
	class Name ## Holder { \
		options::Option<Type>::Ptr d_o ## Name; \
		public : \
			Name ## Holder(options::Section & s) : d_o ## Name ( s.AddOption<Type>(NameInFile,DefaultValue,Comment) ){} \
			inline const Type & Value() const {return d_o ## Name->Value(); }\
			options::Option<Type>::Ptr Option() { return d_o ## Name; }\
	};\
	friend class Name ## Holder; \
	Name ## Holder d_h ## Name;\
	public : \
	inline const Type & Name () const { return d_h ## Name.Value() ; }\
	inline options::Option<Type>::Ptr Name ## Option () { return d_h ## Name.Option() ; }

#define BCPP_ADD_NON_REF_OPTION(Name,NameInFile,Type,DefaultValue,Comment) \
	private :\
	class Name ## Holder { \
		options::Option<Type>::Ptr d_o ## Name; \
		public : \
			Name ## Holder(options::Section & s) : d_o ## Name ( s.AddOption<Type>(NameInFile,DefaultValue,Comment) ){} \
			inline Type  Value() const {return d_o ## Name->Value(); }\
			options::Option<Type>::Ptr Option() { return d_o ## Name; }\
	};\
	friend class Name ## Holder; \
	Name ## Holder d_h ## Name;\
	public : \
	inline Type Name () const { return d_h ## Name.Value() ; }\
	inline options::Option<Type>::Ptr Name ## Option () { return d_h ## Name.Option() ; }

#define BCPP_ADD_SUBSECTION(ConfigSectionType,Name,NameInFile,Comment) \
	private : \
		class Name ## SectionHolder { \
			ConfigSectionType d_ ## Name; \
		public : \
			Name ## SectionHolder(options::Section & s) : d_ ## Name (s.AddSubSection(NameInFile,Comment)) {} \
			inline ConfigSectionType & ConfigSection () { return d_ ## Name ; } \
			inline const ConfigSectionType & ConstConfigSection () const { return d_ ## Name ; } \
		} d_sh ## Name ; \
	public : \
		inline ConfigSectionType & Name () { return d_sh ## Name.ConfigSection(); }; \
		inline const ConfigSectionType & Name () const { return d_sh ## Name.ConstConfigSection(); }



namespace options{

	class ConfigSection {
	public :
		ConfigSection(options::Section::Ptr section) : d_section(section){
		}

		virtual ~ConfigSection(){}

		options::Section & Section() {
			return * d_section;
		}

		const options::Section & Section() const {
			return * d_section;
		}
	protected :
		options::Section::Ptr d_section;
	};

	template <typename SubConfigSection>
	class DynamicConfigSection : public ConfigSection {
	public :
		typedef std::tr1::shared_ptr<SubConfigSection> SubConfigSectionPtr;
		typedef std::map<std::string,SubConfigSectionPtr> ListOfSubsection;

		DynamicConfigSection(options::Section::Ptr section)
			: ConfigSection(section){
			typedef options::Section::SubSectionCreator::Ptr SubSectionCreatorPtr;
			section->SetSubSectionCreator(SubSectionCreatorPtr(new SubSectionCreator(*this)));
		}

		virtual ~DynamicConfigSection(){}

		SubConfigSectionPtr SubSection(const std::string & name){
			typename ListOfSubsection::const_iterator fi = d_subSections.find(name);
			if(d_subSections.end() == fi){
				return SubConfigSectionPtr();
			}
			return fi->second;
		}
		const ListOfSubsection & SubSections(){
			return d_subSections;
		}

	private :
		class SubSectionCreator : public Section::SubSectionCreator {
		public :
			SubSectionCreator(DynamicConfigSection & parent) : d_parent(parent){
			}
			virtual Section::Ptr Create(const std::string & name,
			                            const std::string & comment){
				if(d_parent.d_subSections.find(name) != d_parent.d_subSections.end()){
					throw std::logic_error("Should not happen.");
				}
				options::Section::Ptr newS(new options::Section(name,comment));
				SubConfigSectionPtr scs(new SubConfigSection(newS));

				d_parent.d_subSections.insert(std::make_pair(name,scs));
				return newS;
			}
		private :
			DynamicConfigSection & d_parent;
		};
		ListOfSubsection d_subSections;
	};

/**
 *
 */
class Config  {
public:
	enum Level{
		ALL       = -1,
		SYSTEM    =  0,
		GLOBAL    =  1,
		DIRECTORY =  2
	};
	Config(const std::string & basename);
	virtual ~Config();

	void LoadFile(std::istream & in);
	void LoadFile(const std::string & filename);

	void AppendFile(const std::string & filename);

	void LoadAllFiles(Level l = ALL);

	void Print(std::ostream & out) const;
	void Print(const std::string & filename ) const;

	const std::string & SystemConfigFilename()    const;
	const std::string & GlobalConfigFilename()    const;
	const std::string & DirectoryConfigFilename() const;


protected :
	options::Section & Section(){
		return *d_root;
	}
private :
	std::string CorrectTheBasename(const std::string & basename);
	std::string d_filename;
	std::vector<std::string> d_stdFilenames;


	options::Section::Ptr d_root;

};
} //namespace options

#endif // BIOROB_CPP_OPTIONS_CONFIG_H_
